/* vim:fdm=marker ts=4 et ai
 * {{{
 *
 * (c) by Alexander Neumann <alexander@bumpern.de>
 * Copyright (C) 2007 by Stefan Siegl <stesie@brokenpipe.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * For more information on the GPL, please go to:
 * http://www.gnu.org/copyleft/gpl.html
 }}} */

#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include <util/crc16.h>
#include <string.h>

#include "network.h"
#include "config.h"
#include "eeprom.h"
#include "crc.h"

#include "uip/uip.h"
#include "uip/uip_arp.h"

#ifndef ENC28J60_POLL
#define interrupt_occured() (!(INT_PIN & _BV(INT_PIN_NAME)))
#define wol_interrupt_occured() (!(WOL_PIN & _BV(WOL_PIN_NAME)))
#else
#define interrupt_occured() 0
#define wol_interrupt_occured() 0
#endif

#define BASE_CONFIG ((struct eeprom_config_base_t *)uip_buf)
#define EXT_CONFIG ((struct eeprom_config_ext_t *)uip_buf)

void network_init(void)
/* {{{ */ {

    uip_init();
    uip_arp_init();

    uip_ipaddr_t ipaddr;

    /* load base network settings */

    /* use global network packet buffer for configuration */
    eeprom_read_block(uip_buf, EEPROM_CONFIG_BASE, sizeof(struct eeprom_config_base_t));

    /* checksum */
    uint8_t checksum = crc_checksum(uip_buf, sizeof(struct eeprom_config_base_t) - 1);

    if (checksum != BASE_CONFIG->crc) {
        memcpy_P(uip_ethaddr.addr, PSTR("\xAC\xDE\x48\xFD\x0F\xD0"), 6);
        uip_ipaddr(ipaddr, 10,0,0,5);
        uip_sethostaddr(ipaddr);
        uip_ipaddr(ipaddr, 255,255,255,0);
        uip_setnetmask(ipaddr);
    } else {
        /* load config settings */
        memcpy(uip_ethaddr.addr, &BASE_CONFIG->mac, 6);
        memcpy(&ipaddr, &BASE_CONFIG->ip, 4);
        uip_sethostaddr(ipaddr);
        memcpy(&ipaddr, &BASE_CONFIG->netmask, 4);
        uip_setnetmask(ipaddr);
        memcpy(&ipaddr, &BASE_CONFIG->gateway, 4);
        uip_setdraddr(ipaddr);
    }

    init_enc28j60();
} /* }}} */

void enc28j60_process_interrupts(void)
/* {{{ */ {

    /* also check packet counter, see errata #6 */
#   ifdef ENC28J60_REV4_WORKAROUND
    uint8_t pktcnt = read_control_register(REG_EPKTCNT);
#   endif

    /* if no interrupt occured and less than 5 packets are in the receive
     * buffer, return */
    if ( ! (interrupt_occured()
#   ifdef ENC28J60_REV4_WORKAROUND
                || pktcnt > 5
#   endif
           ) )
        return;

    /* read interrupt register */
    uint8_t EIR = read_control_register(REG_EIR);

    /* clear global interrupt flag */
    bit_field_clear(REG_EIE, _BV(INTIE));

    /* check if some interrupts occured */
    if (EIR != 0) {

        /* check each interrupt flag the interrupt is activated for, and clear it
         * if neccessary */

        /* link change flag */
        if (EIR & _BV(LINKIF)) {
            /* clear interrupt flag */
            read_phy(PHY_PHIR);

            /* read new link state */
            uint8_t link_state = (read_phy(PHY_PHSTAT2) & _BV(LSTAT)) > 0;
        }

        /* packet transmit flag */
        if (EIR & _BV(TXIF)) {
            /* clear flags */
            bit_field_clear(REG_EIR, _BV(TXIF));
            bit_field_clear(REG_ESTAT, _BV(TXABRT) | _BV(LATECOL) );
        }

        /* packet receive flag */
        if (EIR & _BV(PKTIF)) {
            process_packet();
        }

        /* receive error */
        if (EIR & _BV(RXERIF)) {
            bit_field_clear(REG_EIR, _BV(RXERIF));

#ifdef ENC28J60_REV4_WORKAROUND
            init_enc28j60();
#endif
        }

        /* transmit error */
        if (EIR & _BV(TXERIF)) {
            bit_field_clear(REG_EIR, _BV(TXERIF));
        }

    }

    /* set global interrupt flag */
    bit_field_set(REG_EIE, _BV(INTIE));


} /* }}} */

void process_packet(void)
/* {{{ */ {

    /* if there is a packet to process */
    if (read_control_register(REG_EPKTCNT) == 0)
        return;

    /* read next packet pointer */
    set_read_buffer_pointer(enc28j60_next_packet_pointer);
    enc28j60_next_packet_pointer = read_buffer_memory() | (read_buffer_memory() << 8);

    /* read receive status vector */
    struct receive_packet_vector_t rpv;
    uint8_t *p = (uint8_t *)&rpv;

    for (uint8_t i = 0; i < sizeof(struct receive_packet_vector_t); i++)
        *p++ = read_buffer_memory();

    /* decrement rpv received_packet_size by 4, because the 4 byte CRC checksum is counted */
    rpv.received_packet_size -= 4;

    /* check size */
    if (rpv.received_packet_size > MAX_FRAME_LENGTH
            || rpv.received_packet_size < sizeof(struct uip_eth_hdr)
            || rpv.received_packet_size > UIP_BUFSIZE) {
        return;
    }

    /* store packet start pointer for checksum calculation */
    //global_net.packet_start_pointer = get_read_buffer_pointer();

    /* read packet */
    p = uip_buf;
    for (uint16_t i = 0; i < rpv.received_packet_size; i++)
        *p++ = read_buffer_memory();

    uip_len = rpv.received_packet_size;

    /* process packet */
    struct uip_eth_hdr *packet = (struct uip_eth_hdr *)&uip_buf;
    switch (ntohs(packet->type)) {

        /* process arp packet */
        case UIP_ETHTYPE_ARP:
            uip_arp_arpin();

            /* if there is a packet to send, send it now */
            if (uip_len > 0)
                transmit_packet();

            break;

        /* process ip packet */
        case UIP_ETHTYPE_IP:
            uip_arp_ipin();
            uip_input();

            /* if there is a packet to send, send it now */
            if (uip_len > 0) {

                /* check if an arp request has to be send */
                uip_arp_out();
                transmit_packet();
            }

            break;
    }

    /* advance receive read pointer, ensuring that an odd value is programmed
     * (next_receive_packet_pointer is always even), see errata #13 */
    if ( (enc28j60_next_packet_pointer - 1) < RXBUFFER_START
            || (enc28j60_next_packet_pointer - 1) > RXBUFFER_END) {

        write_control_register(REG_ERXRDPTL, LO8(RXBUFFER_END));
        write_control_register(REG_ERXRDPTH, HI8(RXBUFFER_END));

    } else {

        write_control_register(REG_ERXRDPTL, LO8(enc28j60_next_packet_pointer - 1));
        write_control_register(REG_ERXRDPTH, HI8(enc28j60_next_packet_pointer - 1));

    }

    /* decrement packet counter */
    bit_field_set(REG_ECON2, _BV(PKTDEC));

} /* }}} */

void transmit_packet(void)
/* {{{ */ {

    /* wait for any transmits to end, with timeout */
    uint8_t timeout = 100;
    while (read_control_register(REG_ECON1) & _BV(ECON1_TXRTS) && timeout-- > 0);

    if (timeout == 0) {

    } else {

        uint16_t start_pointer = TXBUFFER_START;

        /* set send control registers */
        write_control_register(REG_ETXSTL, LO8(start_pointer));
        write_control_register(REG_ETXSTH, HI8(start_pointer));

        write_control_register(REG_ETXNDL, LO8(start_pointer + uip_len));
        write_control_register(REG_ETXNDH, HI8(start_pointer + uip_len));

        /* set pointer to beginning of tx buffer */
        set_write_buffer_pointer(start_pointer);

        /* write override byte */
        write_buffer_memory(0);

        /* write data */
        for (uint16_t i = 0; i < uip_len; i++)
            write_buffer_memory(uip_buf[i]);

#       ifdef ENC28J60_REV4_WORKAROUND
        /* reset transmit hardware, see errata #12 */
        bit_field_set(REG_ECON1, _BV(ECON1_TXRST));
        bit_field_clear(REG_ECON1, _BV(ECON1_TXRST));
#       endif

        /* transmit packet */
        bit_field_set(REG_ECON1, _BV(ECON1_TXRTS));

    }

} /* }}} */
